"""
    Dummy conftest.py for osw.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""

import pytest


def pytest_addoption(parser):
    parser.addoption("--wiki_domain", action="store")
    parser.addoption("--wiki_username", action="store")
    parser.addoption("--wiki_password", action="store")


@pytest.fixture(scope="session")
def wiki_domain(request):
    value = request.config.option.wiki_domain
    if value is None:
        pytest.skip()
    return value


@pytest.fixture(scope="session")
def wiki_username(request):
    value = request.config.option.wiki_username
    if value is None:
        pytest.skip()
    return value


@pytest.fixture(scope="session")
def wiki_password(request):
    value = request.config.option.wiki_password
    if value is None:
        pytest.skip()
    return value



import mwclient

# from osw.core import OSW
# from osw.wtsite import WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass


def test_login(wiki_domain, wiki_username, wiki_password):
    site = mwclient.Site(host=wiki_domain)
    site.login(username=wiki_username, password=wiki_password)
    # wtsite = WtSite(site)
    # osw = OSW(site=wtsite)



import os

# import osw.controller.entity as controller
import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

wtsite.enable_cache()  # skip downloads on second request

# osw.fetch_schema() #this will load the current entity schema from the OSW instance. You may have to re-run the script to get the updated schema extension. Requires 'pip install datamodel-code-generator'
if False:
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW02590972aeba46d7864ed492c0c11384", mode="replace"
        )
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSWacdb001c926c46b998af3e645d36b13f", mode="append"
        )
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="Category:OSW51ad0d1716254c77a2b7a03217f23b43", mode="append"
        )
    )

# #explicit load host and server
# host = osw.load_entity("Item:OSWb8625aedf4074b72a4fa9bcb11694e4d")
# server = osw.load_entity("Item:OSW536e1417a0eb427887aeb267f47442fb")

db = osw.load_entity("Item:OSW837099bdc67d4b7cb5ffcce713a7648b")
db = db.cast(model.Database)

# #load host and server from references
# server = osw.load_entity(db.db_server)
# server = server.cast(model.DatabaseServer)
# host = osw.load_entity(server.host)
# host = host.cast(model.Host)

# #load host and server from semantic queries
server_title = wtsite.semantic_search(
    f"[[-HasDbServer::Item:{osw.get_osw_id(db.uuid)}]]"
)
server = osw.load_entity(server_title[0])

host_title = wtsite.semantic_search(f"[[-HasHost::Item:{osw.get_osw_id(server.uuid)}]]")
host = osw.load_entity(host_title[0])


connect_str = f"{host.network_domain[0]}:{server.network_port[0]}/{db.db_name}"
print(connect_str)

# cast into controller and execute function
# db = db.cast(controller.DbController)
# print(db.create_table())



import os
from importlib import reload

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

# osw.fetch_schema() #this will load the current entity schema from the OSW instance. You may have to re-run the script to get the updated schema extension. Requires 'pip install datamodel-code-generator'
osw.fetch_schema(OSW.FetchSchemaParam(schema_title="Category:Hardware", mode="replace"))
osw.fetch_schema(
    OSW.FetchSchemaParam(
        schema_title="Category:OSWf9742b744112486ea2a5d8afee3e2538", mode="append"
    )
)
reload(model)



import osw.model.entity as model


class Entity(model.Entity):
    def explain(self):
        print(f"Entity with label '{str(self.label[0].text)}'")


# model.Entity = Entity


class Hardware(model.Hardware, Entity):
    def run(self):
        self.explain()
        print(" is running")


# model.Hardware = Hardware


class DbController(model.Database):
    def create_table(self):
        connection_str = self.db_server
        return connection_str



# generated by datamodel-codegen:
#   filename:  Item.json
#   timestamp: 2023-01-23T06:27:29+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, List, Optional
from uuid import UUID

from pydantic import BaseModel, Field
from typing_extensions import Literal


class LangCode(Enum):
    en = "en"
    de = "de"


from typing import TYPE_CHECKING, Type, TypeVar
from uuid import uuid4

if TYPE_CHECKING:
    from dataclasses import dataclass as _basemodel_decorator
else:
    _basemodel_decorator = lambda x: x  # noqa: E731


T = TypeVar("T", bound=BaseModel)


class OswBaseModel(BaseModel):
    def full_dict(self, **kwargs):  # extent BaseClass export function
        d = super().dict(**kwargs)
        for key in ("_osl_template", "_osl_footer"):
            if hasattr(self, key):
                d[key] = getattr(
                    self, key
                )  # include selected private properites. note: private properties are not considered as discriminator
        return d

    def cast(self, cls: Type[T]) -> T:
        return cls(**self.dict())

    def cast_none_to_default(self, cls: Type[T]) -> T:
        """Casting self into target class. If the passed attribute is None or solely
        includes None values, the attribute is not passed to the instance of the
        target class, which will then fall back to the default."""

        def test_if_empty_list_or_none(obj) -> bool:
            if obj is None:
                return True
            elif isinstance(obj, list):
                if len(obj) == 0:
                    return True
                elif len([item for item in obj if item is not None]) == 0:
                    return True
            return False

        return cls(
            **{
                k: v
                for k, v in self.dict().items()
                if not test_if_empty_list_or_none(v)
            }
        )


@_basemodel_decorator
class Label(OswBaseModel):
    text: str = Field(..., title="Text")
    lang: Optional[LangCode] = Field("en", title="Lang code")


class LangCode1(Enum):
    en = "en"
    de = "de"


@_basemodel_decorator
class Description(OswBaseModel):
    text: str = Field(..., title="Text")
    lang: Optional[LangCode1] = Field("en", title="Lang code")


@_basemodel_decorator
class Statement(OswBaseModel):
    osl_template: Optional[Literal["OslTemplate:Statement"]] = "OslTemplate:Statement"
    label: Optional[Label] = Field(None, title="Label")
    """
    Human readable name
    """
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    subject: Optional[str] = None
    predicate: str
    object: Optional[str] = None
    substatements: Optional[List[Statement]] = Field(None, title="Substatements")


@_basemodel_decorator
class Entity(OswBaseModel):
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    label: Label = Field(..., title="Label")
    """
    Human readable name
    """
    query_label: Optional[str] = Field(None, title="Label")
    additional_labels: Optional[List[Label]] = Field(None, title="Additional Labels")
    description: Optional[List[Description]] = Field(None, title="Description")
    image: Optional[str] = Field(None, title="Image")
    statements: Optional[List[Statement]] = Field(None, title="Statements")
    extensions: Optional[List[Any]] = Field(None, title="Extensions")


class Item(Entity):
    type: Optional[List[str]] = Field(
        ["Category:Item"], min_length=1, title="Types/Categories"
    )


Statement.update_forward_refs()



"""
# todo: docstring erzeugen

verwende: beispiel create page package
https://github.com/OpenSemanticLab/osw-python/blob/main/examples/create_page_package.py
in wtsite neuen parameter hinzufügen --> label als file name:
dumpy_empty_slots = False
https://github.com/OpenSemanticLab/osw-python/blob/main/src/osw/wtsite.py#L112
"""
import json
import os

# import yaml
from pathlib import Path
from typing import Union

import PySimpleGUI as psg
from numpy import array as np_array

import osw.model.page_package as package
from osw.core import OSW
from osw.wiki_tools import read_domains_from_credentials_file
from osw.wtsite import SLOTS, WtPage, WtSite

# Definition of constants
GUI_THEME = "reddit"
DEBUG = True  # Set to True to get an output element in the GUI
DUMP_EMPTY_SLOTS_DEFAULT = True
SEL_INDICES_DEFAULT = [
    0,  # "main",
    3,  # "jsondata",
    4,  # "jsonschema",
    6,  # "header_template",
]
SLOTS_TO_UPLOAD_DEFAULT = np_array(list(SLOTS.keys()))[SEL_INDICES_DEFAULT].tolist()
TARGET_PAGE_DEFAULT = "https://wiki-dev.open-semantic-lab.org/wiki/Main_Page"
LWD_DEFAULT = Path(os.getcwd()).parent / "data"
SETTINGS_FILE_PATH_DEFAULT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "settings.json"
)
CREDENTIALS_FILE_PATH_DEFAULT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)


# Definition of functions
def str_or_none(value):
    if str(value) == "" or value is None:
        return str("None")
    else:
        return str(value)


def create_config_from_setting(settings_: dict):
    config_ = WtPage.PageDumpConfig(
        target_dir=settings_["local_working_directory"],
        dump_empty_slots=settings_["dump_empty_slots"],
        page_name_as_filename=settings_["page_name_as_filename"],
    )
    return config_


def create_page_package(
    full_page_name_str,
    wtsite_inst: WtSite,
    dump_config_inst: WtPage.PageDumpConfig,
    label_str: str = None,
    top_level: str = None,
    sub_level: str = "content",
    author: Union[str, list] = "Open Semantic World",
):
    if top_level is None:
        top_level = full_page_name_str.split(":")[-1]
    package_repo_org = "OpenSemanticWorld-Packages"
    package_repo = f"{top_level}.{sub_level}"
    package_id = f"{top_level}.{sub_level}"
    package_name = top_level
    package_subdir = sub_level
    package_branch = "deleteme"
    publisher = "Open Semantic World"
    working_dir = dump_config_inst.target_dir
    if dump_config_inst.page_name_as_filename and label_str is not None:
        target_dir = os.path.join(working_dir, label_str)
    else:
        target_dir = os.path.join(working_dir, top_level)
    if isinstance(author, list):
        author_list = author
    elif isinstance(author, str):
        author_list = [author]
    else:
        author_list = [""]

    bundle = package.PagePackageBundle(
        publisher=publisher,
        author=author_list,
        language="en",
        publisherURL=f"https://github.com/{package_repo_org}/{package_repo}",
        packages={
            f"{package_name}": package.PagePackage(
                globalID=f"{package_id}",
                label=package_name,
                version="0.0.1",
                description="Created by the GUI for local editing of OSW pages",
                baseURL=f"https://raw.githubusercontent.com/{package_repo_org}/"
                f"{package_repo}/{package_branch}/{package_subdir}/",
            )
        },
    )
    wtsite_inst.create_page_package(
        config=package.PagePackageConfig(
            name=package_name,
            config_path=os.path.join(target_dir, "packages.json"),
            content_path=os.path.join(target_dir, package_subdir),
            bundle=bundle,
            titles=[full_page_name_str],
        ),
        dump_config=dump_config_inst,
        debug=False,
    )


# Predefining some variables before execution
settings_file_path = SETTINGS_FILE_PATH_DEFAULT
if os.path.exists(settings_file_path):
    with open(settings_file_path, "r") as f:
        settings = json.load(f)
    settings_read_from_file = True
else:
    settings = {
        "credentials_file_path": str(CREDENTIALS_FILE_PATH_DEFAULT),
        "local_working_directory": str(LWD_DEFAULT),
        "settings_file_path": str(SETTINGS_FILE_PATH_DEFAULT),
        "target_page": TARGET_PAGE_DEFAULT,
        "dump_empty_slots": DUMP_EMPTY_SLOTS_DEFAULT,
        "page_name_as_filename": True,
        "slots_to_upload": SLOTS_TO_UPLOAD_DEFAULT,
        "domain": "",
    }
    settings_read_from_file = False

domains, accounts = read_domains_from_credentials_file(
    settings["credentials_file_path"]
)
domain = domains[0]
if settings_read_from_file:
    settings["domain"] = domain

wtsite_obj = WtSite.from_domain(
    domain=domains[0], password_file="", credentials=accounts[domains[0]]
)
osw_obj = OSW(site=wtsite_obj)

full_page_name = settings["target_page"].split("/")[-1].replace("_", " ")
page = wtsite_obj.get_WtPage(full_page_name)
label_set = False
label = None

# ----- GUI Definition -----
# Setting the theme of the GUI
psg.theme(GUI_THEME)
# Settings
settings_layout = [
    [psg.Text("Settings", font=("Helvetica", 20))],
    [
        psg.Column(
            [
                [
                    psg.Text("Settings file"),
                ],
                [
                    psg.Text("Credentials file"),
                ],
                [
                    psg.Text("Local working directory"),
                ],
            ]
        ),
        psg.Column(
            [
                [
                    psg.In(
                        size=(50, 1),
                        enable_events=True,
                        key="-SETTINGS-",
                        default_text=settings["settings_file_path"],
                    ),
                    psg.FileBrowse(button_text="Browse", key="-BROWSE_SETTINGS-"),
                    psg.Button("Load", key="-LOAD_SETTINGS-"),
                    psg.Button("Save", key="-SAVE_SETTINGS-"),
                ],
                [
                    psg.In(
                        size=(50, 1),
                        enable_events=True,
                        key="-CREDENTIALS-",
                        default_text=settings["credentials_file_path"],
                    ),
                    psg.FileBrowse(button_text="Browse", key="-BROWSE_CREDENTIALS-"),
                ],
                [
                    psg.In(
                        size=(50, 1),
                        enable_events=True,
                        key="-LWD-",
                        default_text=settings["local_working_directory"],
                    ),
                    psg.FolderBrowse(button_text="Browse", key="-BROWSE_LWD-"),
                ],
            ]
        ),
    ],
    [psg.HSeparator()],
]
# Actions
actions_layout = [
    # Actions
    [psg.Text("Actions", font=("Helvetica", 20))],
    # Target OSW instance
    [psg.Text("Target OSW instance", font=("Helvetica", 16))],
    [psg.Text("List of domains is read from accounts.pwd.yaml!")],
    [psg.Combo(domains, default_value=domains[0], key="-DOMAIN-", enable_events=True)],
    # Target page
    [psg.Text("Target page", font=("Helvetica", 16))],
    [
        psg.Column(
            [
                [psg.Text("Address, within selected OSW instance")],
                [psg.Text("First label of the OSW page:")],
            ]
        ),
        psg.Column(
            [
                [
                    psg.InputText(
                        size=(50, 1),
                        default_text=settings["target_page"],
                        key="-ADDRESS-",
                    ),
                    psg.Button("Load page"),
                ],
                [
                    # A display element that will show the label of the OSW page
                    psg.Multiline(size=(50, 1), key="-LABEL-", no_scrollbar=True)
                ],
            ]
        ),
    ],
    # [
    #     psg.HSeparator()
    # ],
    # Slots to download
    [psg.Text("Slots to download", font=("Helvetica", 16))],
    [
        psg.Column(
            [
                [
                    psg.Radio(
                        "Include empty slots",
                        group_id="-RADIO1-",
                        default=settings["dump_empty_slots"],
                        key="-INC_EMPTY-",
                        enable_events=True,
                    )
                ],
                [
                    psg.Radio(
                        "Exclude empty slots",
                        group_id="-RADIO1-",
                        default=(not settings["dump_empty_slots"]),
                        key="-EXC_EMPTY-",
                        enable_events=True,
                    )
                ],
                [psg.Button("Download selected", key="-DL-")],
                [psg.Multiline(size=(20, 1), key="-DL_RES-", no_scrollbar=True)],
            ]
        ),
    ],
    # [
    #     psg.HSeparator()
    # ],
    # Slots to upload
    [psg.Text("Slots to upload", font=("Helvetica", 16))],
    [
        psg.Column(
            [
                [psg.Button("(De)Select all", key="-UL_SEL-")],
                [
                    psg.Listbox(
                        values=SLOTS.keys(),
                        default_values=settings["slots_to_upload"],
                        size=(20, 10),
                        key="-UL_LIST-",
                        select_mode=psg.LISTBOX_SELECT_MODE_MULTIPLE,
                        enable_events=True,
                        no_scrollbar=True,
                    )
                ],
                [psg.Button("Upload selected", key="-UL-")],
                [psg.Multiline(size=(20, 1), key="-UL_RES-", no_scrollbar=True)],
            ]
        )
    ],
]
output_layout = [
    [psg.HSeparator()],
    # Output
    [psg.Text("Debug output", font=("Helvetica", 20))],
    [psg.Multiline(key="-OUTPUT-", size=(100, 10))],
]
# The full layout
layout = settings_layout + actions_layout
if DEBUG:
    layout += output_layout
# Create the window
window = psg.Window("OSW local editing GUI", layout)
# Create an event loop
while True:
    event, values = window.read()
    # End program if user closes window or
    # presses the OK button
    if event == "OK" or event == psg.WIN_CLOSED:
        break
    elif event == "-SETTINGS-":
        settings["settings_file_path"] = values["-SETTINGS-"]
    elif event == "-LOAD_SETTINGS-":
        # update settings
        with open(settings["settings_file_path"], "r") as f:
            settings = json.load(f)
        # update GUI
        window["-CREDENTIALS-"].update(settings["credentials_file_path"]),
        window["-LWD-"].update(settings["local_working_directory"])
        window["-DOMAIN-"].update(settings["domain"])
        window["-ADDRESS-"].update(settings["target_page"])
        window["-INC_EMPTY-"].update(settings["dump_empty_slots"])
        window["-EXC_EMPTY-"].update(not settings["dump_empty_slots"])
        indices = [
            i for i, x in enumerate(SLOTS.keys()) if x in settings["slots_to_upload"]
        ]
        window["-UL_LIST-"].update(set_to_index=indices)
        settings_read_from_file = True
    elif event == "-SAVE_SETTINGS-":
        with open(settings["settings_file_path"], "w") as f:
            json.dump(settings, f, indent=4)
    elif event == "-CREDENTIALS-":
        settings["credentials_file_path"] = values["-CREDENTIALS-"]
        domains, accounts = read_domains_from_credentials_file(
            settings["credentials_file_path"]
        )
        window["-DOMAIN-"].update(values=domains)
    elif event == "-LWD-":
        settings["local_working_directory"] = values["-LWD-"]
    elif event == "-DOMAIN-":
        settings["domain"] = values["-DOMAIN-"]
        domain = settings["domain"].split("//")[-1]
        wtsite_obj = WtSite.from_domain(
            domain=settings["domain"],
            password_file="",
            credentials=accounts[settings["domain"]],
        )
        osw_obj = OSW(site=wtsite_obj)
    elif event == "Load page":
        full_page_name = values["-ADDRESS-"].split("/")[-1].replace("_", " ")
        if (values["-ADDRESS-"].find("/wiki/") != -1) or (
            values["-ADDRESS-"].find("/w/") != -1
        ):
            settings["target_page"] = values["-ADDRESS-"]
        else:
            settings["target_page"] = "https://" + domain + "/wiki/" + full_page_name
        if values["-ADDRESS-"].find(settings["domain"]) == -1:
            window["-LABEL-"].update("Page not on selected domain!")
            label_set = False
        else:
            # use connection
            page = wtsite_obj.get_WtPage(full_page_name)
            if page.exists:
                jsondata = page.get_slot_content("jsondata")
                if jsondata is None:
                    window["-LABEL-"].update("Slot 'jsondata' is empty!")
                    label_set = False
                else:
                    label = jsondata["label"][0]["text"]
                    window["-LABEL-"].update(label)
                    label_set = True
            else:
                window["-LABEL-"].update("Page does not exist!")
                label_set = False
    elif event == "-EXC_EMPTY-" or event == "-INC_EMPTY-":
        settings["dump_empty_slots"] = values["-INC_EMPTY-"]
    elif event == "-DL-":
        if label_set:
            dump_config = create_config_from_setting(settings)
            # _ = page.dump(dump_config)
            create_page_package(
                full_page_name_str=full_page_name,
                wtsite_inst=wtsite_obj,
                dump_config_inst=dump_config,
                label_str=label,
                author=accounts[domains[0]]["username"],
            )
            window["-DL_RES-"].update("Slots downloaded!")
        else:
            window["-DL_RES-"].update("No page loaded!")
    elif event == "-UL_SEL-":
        indices = list(window["-UL_LIST-"].get_indexes())
        # All selected: deselect all
        if indices == list(range(len(SLOTS.keys()))):
            window["-UL_LIST-"].update(set_to_index=[])
        else:
            window["-UL_LIST-"].update(set_to_index=list(range(len(SLOTS.keys()))))
    elif event == "-UL_LIST-":
        indices = list(window["-UL_LIST-"].get_indexes())
        settings["slots_to_upload"] = np_array(list(SLOTS.keys()))[indices].tolist()
    elif event == "-UL-":
        pass
        # Success:
        # window["-UL_RES-"].update("Slots uploaded!")
        window["-UL_RES-"].update("Not yet implemented!")
        # todo: no slot selected

    # Some debugging output functionality
    if DEBUG:
        values_str = "\n".join(
            [
                f"{key}: {str_or_none(value)}"
                for key, value in values.items()
                if key != "-OUTPUT-"
            ]
        )
        listbox_selected_items_str = "\n".join(settings["slots_to_upload"])

        window["-OUTPUT-"].update(
            f"Last event: {event}\n"
            f"-----------\n"
            f"Values:\n"
            f"-------\n"
            f"{values_str}"
            f"\nListbox values:\n"
            f"{listbox_selected_items_str}"
        )

window.close()



# extents mwclient.site

import json
import os
import shutil
from datetime import datetime
from pathlib import Path
from pprint import pprint
from time import sleep
from typing import Dict, Optional, Union

import mwclient
from jsonpath_ng.ext import parse
from pydantic import BaseModel

import osw.model.page_package as package
import osw.wiki_tools as wt
from osw.model.entity import _basemodel_decorator

# Definition of constants
SLOTS = {
    "main": {"content_model": "wikitext", "content_template": ""},
    "header": {"content_model": "wikitext", "content_template": ""},
    "footer": {"content_model": "wikitext", "content_template": ""},
    "jsondata": {"content_model": "json", "content_template": ""},
    "jsonschema": {"content_model": "json", "content_template": ""},
    "template": {"content_model": "wikitext", "content_template": ""},
    "header_template": {"content_model": "wikitext", "content_template": ""},
    "footer_template": {"content_model": "wikitext", "content_template": ""},
    "data_template": {"content_model": "wikitext", "content_template": ""},
    "schema_template": {"content_model": "wikitext", "content_template": ""},
}


class WtSite:
    def __init__(self, site: mwclient.Site = None):
        if site:
            self._site = site
        else:
            raise ValueError("Parameter 'site' is None")
        self._page_cache = {}
        self._cache_enabled = False

    @classmethod
    def from_domain(
        cls,
        domain: str = None,
        password_file: Union[str, Path] = None,
        credentials: dict = None,
    ):
        if credentials is None:
            site = wt.create_site_object(domain, password_file)
        else:
            site = wt.create_site_object(domain, "", credentials)
        return cls(site)

    @classmethod
    def from_credentials(
        cls,
        credentials: Union[Dict[str, Dict[str, str]], str, Path],
        key: Union[str, int] = 0,
    ):
        """

        Parameters
        ----------
        credentials:
            A dictionary of credentials or a path to a credentials file, expected to
            contain keys 'username' and 'password'.
        key:
            The key of the credentials-dictionary (domain) or the index of the key.
            By default, the first key is used.

        Returns
        -------

        """

        if isinstance(credentials, str) or isinstance(credentials, Path):
            domains, accounts = wt.read_domains_from_credentials_file(credentials)
        else:
            accounts = credentials
        if isinstance(key, int):
            _domain = list(accounts.keys())[key]
        else:
            _domain = key
        _domain = _domain
        _credentials: Dict = accounts[_domain]
        # todo: research why nested typing doesn't work as expected fo the dictionary
        site = wt.create_site_object(_domain, "", _credentials)
        return cls(site)

    def get_WtPage(self, title: str = None):
        retry = 0
        max_retry = 5
        page = None
        while retry < max_retry:
            try:
                page = self._get_WtPage(title)
                break
            except Exception as e:
                print(e)
                if retry < max_retry:
                    retry += 1
                    print(f"Page load failed. Retry ({retry}/{max_retry})")
                    sleep(5)
        self._clear_cookies()
        return page

    def _get_WtPage(self, title: str = None):

        if self._cache_enabled and title in self._page_cache:
            return self._page_cache[title]
        else:
            wtpage = WtPage(self, title)
            if self._cache_enabled:
                self._page_cache[title] = wtpage

        return wtpage

    def enable_cache(self):
        self._cache_enabled = True

    def disable_cache(self):
        self._cache_enabled = False

    def get_cache_enabled(self):
        return self._cache_enabled

    def clear_cache(self):
        del self._page_cache
        self._page_cache = {}

    def _clear_cookies(self):
        # see https://github.com/mwclient/mwclient/issues/221
        for cookie in self._site.connection.cookies:
            if "PostEditRevision" in cookie.name:
                self._site.connection.cookies.clear(
                    cookie.domain, cookie.path, cookie.name
                )

    def prefix_search(self, text):
        return wt.prefix_search(self._site, text)

    def semantic_search(self, query):
        return wt.semantic_search(self._site, query)

    def modify_search_results(
        self,
        mode: str,
        query: str,
        modify_page,
        limit=None,
        comment=None,
        log=False,
        dryrun=False,
    ):
        titles = []
        if mode == "prefix":
            titles = wt.prefix_search(self._site, query)
        elif mode == "semantic":
            titles = wt.semantic_search(self._site, query)
        if limit:
            titles = titles[0:limit]
        if log:
            print(f"Found: {titles}")
        for title in titles:
            wtpage = self.get_WtPage(title)
            modify_page(wtpage)
            if log:
                print(f"\n======= {title} =======")
                print(wtpage._content)
                for slot in wtpage._slots:
                    content = wtpage.get_slot_content(slot)
                    # if isinstance(content, dict): content = json.dumps(content)
                    print(f"   ==== {title}:{slot} ====   ")
                    pprint(content)
                    print("\n")
            if not dryrun:
                wtpage.edit(comment)

    def create_page_package(
        self,
        config: package.PagePackageConfig,
        dump_config: "WtPage.PageDumpConfig" = None,
        debug: bool = True,
    ):
        """Create a page package, which is a locally stored collection of wiki pages
        and their slots, based on a configuration
        object.
        """
        # Clear the content directory
        try:
            if debug:
                print(f"Delete dir '{config.content_path}'")
            if os.path.exists(config.content_path):
                shutil.rmtree(config.content_path)
        except OSError as e:
            if debug:
                print("Error: %s - %s." % (e.filename, e.strerror))
        # Create a dump config
        if dump_config is None:
            dump_config = WtPage.PageDumpConfig(
                target_dir=config.content_path,
                skip_slot_suffix_for_main=config.skip_slot_suffix_for_main,
            )
        else:
            dump_config.target_dir = config.content_path
            dump_config.skip_slot_suffix_for_main = config.skip_slot_suffix_for_main

        bundle = config.bundle  # type: package.PagePackageBundle
        if config.name not in bundle.packages:
            print(f"Error: package {config.name} does not exist in bundle")
            return
        if not bundle.packages[config.name].pages:
            bundle.packages[config.name].pages = []
        for title in config.titles:
            page = self.get_WtPage(title)
            bundle.packages[config.name].pages.append(page.dump(dump_config))
            if config.include_files:
                for file in page._page.images():
                    file_page = self.get_WtPage(file.name)
                    bundle.packages[config.name].pages.append(
                        file_page.dump(dump_config)
                    )

        content = bundle.json(exclude_none=True, indent=4)
        # This will create the JSON (e.g., package.json) with the PagePackageConfig,
        #  which contains the PagePackageBundle
        file_name = f"{config.config_path}"
        with open(file_name, "w") as f:
            f.write(content)

    # todo: implement

    def upload_page_package(self):
        pass


class WtPage:
    def __init__(self, wtSite: WtSite = None, title: str = None):
        self.wtSite = wtSite
        self.title = title

        self._page = wtSite._site.pages[self.title]
        self.exists = self._page.exists
        self._original_content = ""
        self._content = ""
        self.changed = False
        self._dict = []  # todo: named dict but is of type list
        self._slots = {"main": ""}
        self._slots_changed = {"main": False}
        self._content_model = {"main": "wikitext"}

        if self.exists:
            self._original_content = self._page.text()
            self._content = self._original_content
            self._dict = wt.create_flat_content_structure_from_wikitext(
                self._content, array_mode="only_multiple"
            )
            # multi content revisions
            rev = wtSite._site.api(
                "query",
                prop="revisions",
                titles=title,
                rvprop="ids|timestamp|flags|comment|user|content|contentmodel|roles|slotsize",
                rvslots="*",
                rvlimit="1",
                format="json",
            )
            for page_id in rev["query"]["pages"]:
                page = rev["query"]["pages"][page_id]
                if page["title"] == title:
                    for revision in page["revisions"]:
                        self._current_revision = revision
                        for slot_key in revision["slots"]:
                            self._slots[slot_key] = revision["slots"][slot_key]["*"]
                            self._content_model[slot_key] = revision["slots"][slot_key][
                                "contentmodel"
                            ]
                            self._slots_changed[slot_key] = False
                            if self._content_model[slot_key] == "json":
                                self._slots[slot_key] = json.loads(
                                    self._slots[slot_key]
                                )
                    # todo: set content for slots not in revision["slots"] (use
                    #  SLOTS) --> create empty slots

    def create_slot(self, slot_key, content_model):
        self._slots[slot_key] = None
        self._slots_changed[slot_key] = False
        self._content_model[slot_key] = content_model

    def get_content(self):
        return self._content

    def get_slot_content(self, slot_key):
        if slot_key not in self._slots:
            return None
        return self._slots[slot_key]

    def get_slot_content_model(self, slot_key):
        if slot_key not in self._slots:
            return None
        return self._content_model[slot_key]

    def set_content(self, content):
        self._content = content
        self.changed = True

    def set_slot_content(self, slot_key, content):
        # todo: get slot content model from page / constant
        if slot_key not in self._slots:
            content_model = "json"
            if type(content) == str:
                content_model = "wikitext"
            self.create_slot(slot_key, content_model)
        if content != self._slots[slot_key]:
            self._slots_changed[slot_key] = True
        self._slots[slot_key] = content

    def get_url(self) -> str:
        return "https://" + self.wtSite._site.host + "/wiki/" + self.title

    def is_file_page(self) -> bool:
        """Checks if this page is a file page (containing an image, pdf, etc.)

        Returns
        -------
            true if this page is a file page, else false
        """
        return self.title.startswith("File:")

    def append_template(self, template_name: str = None, template_params: dict = None):
        self._dict.append({template_name: template_params})
        return self

    def append_text(self, text):
        self._dict.append(text)
        return self

    def get_value(self, jsonpath):
        jsonpath_expr = parse(jsonpath)
        res = []
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        for match in jsonpath_expr.find(d):
            res.append(match.value)
        return res

    def update_dict(combined: dict, update: dict) -> None:
        for k, v in update.items():
            if isinstance(v, dict):
                WtPage.combine_into(v, combined.setdefault(k, {}))
            else:
                combined[k] = v

    def set_value(self, jsonpath_match, value, replace=False):
        jsonpath_expr = parse(jsonpath_match)
        d = dict(
            zip(range(len(self._dict)), self._dict)
        )  # convert list to dict with index
        # if create: jsonpath_expr.update_or_create(d, value)
        # else: jsonpath_expr.update(d, value)
        matches = jsonpath_expr.find(d)
        for match in matches:
            print(match.full_path)
            # pprint(value)
            if not replace:
                WtPage.update_dict(match.value, value)
                value = match.value
            # pprint(value)
            match.full_path.update_or_create(d, value)
        self._dict = list(d.values())  # convert dict with index to list
        return self

    def update_content(self):
        self._content = wt.get_wikitext_from_flat_content_structure(self._dict)
        self.changed = self._original_content != self._content
        return self

    def edit(self, comment: str = None, mode="action-multislot"):
        """creates / updates the content of all page slots in the wiki site

        Parameters
        ----------
        comment, optional
            edit comment for the page history, by default None
        mode, optional
            single API call ('action-multislot') or multiple ('action-singleslot'), by default 'action-multislot' (faster)
        """
        retry = 0
        max_retry = 5
        while retry < max_retry:
            try:
                return self._edit(comment, mode)
            except Exception as e:
                print(e)
                if retry < max_retry:
                    retry += 1
                    print(f"Page edit failed. Retry ({retry}/{max_retry})")
                    sleep(5)

    def _edit(self, comment: str = None, mode="action-multislot"):
        if not comment:
            comment = "[bot] update of page content"
        if self.changed:
            self._page.edit(self._content, comment)  # legacy mode
        if mode == "action-multislot":
            params = {}
            changed = False
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    changed = True
                    self._slots_changed[slot_key] = False
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        content = json.dumps(content)
                    params["slot_" + slot_key] = content
            if changed:
                self.wtSite._site.api(
                    "editslots",
                    token=self.wtSite._site.get_token("csrf"),
                    title=self.title,
                    summary=comment,
                    **params,
                )
                self.wtSite._clear_cookies()

        else:
            for slot_key in self._slots:
                if self._slots_changed[slot_key]:
                    content = self._slots[slot_key]
                    if self._content_model[slot_key] == "json":
                        content = json.dumps(content)
                    self.wtSite._site.api(
                        "editslot",
                        token=self.wtSite._site.get_token("csrf"),
                        title=self.title,
                        slot=slot_key,
                        text=content,
                        summary=comment,
                    )
                    self._slots_changed[slot_key] = False

    def delete(self, comment: str = None):
        self._page.delete(comment)

    def move(self, new_title: str, comment: str = None, redirect=True):
        if new_title != self.title:
            print(f"move '{self.title}' to '{new_title}'")
            self._page.move(
                new_title=new_title, reason=comment, no_redirect=not redirect
            )
            self.title = new_title

    def get_last_changed_time(self):
        return datetime.fromisoformat(
            self._current_revision["timestamp"].replace("Z", "+00:00")
        )

    @_basemodel_decorator
    class PageDumpConfig(BaseModel):
        target_dir: Union[str, Path]
        namespace_as_folder: Optional[bool] = True
        skip_slot_suffix_for_main: Optional[bool] = False
        dump_empty_slots: Optional[bool] = False
        page_name_as_filename: Optional[bool] = False

        class Config:
            arbitrary_types_allowed = True  # necessary to allow e.g. np.array as type

    def dump(self, config: PageDumpConfig):
        page_name = self.title.split(":")[-1]
        if ":" in self.title:
            namespace = self.title.split(":")[0]
        else:
            namespace = "Main"
        namespace_const = "NS_" + namespace.upper()
        # namespace_id = self._page.namespace

        package_page = package.PagePackagePage(
            name=page_name, namespace=namespace_const, slots={}
        )
        name_in_json_data = False
        if "jsondata" in self._slots and "name" in self._slots["jsondata"]:
            package_page.label = self._slots["jsondata"]["name"]
            name_in_json_data = True

        if config.page_name_as_filename and name_in_json_data:
            # Use name from jsondata as filename for the dump
            dump_name = self._slots["jsondata"]["name"]
        else:
            # Use page name as filename for the dump
            dump_name = page_name

        tar_dir = config.target_dir
        path_prefix = ""
        if config.namespace_as_folder:
            tar_dir = os.path.join(tar_dir, namespace)
            path_prefix = namespace + "/"
        # if not os.path.exists(dir):
        #    os.makedirs(dir)

        def save_to_file(file_path__, content__):
            if not os.path.exists(os.path.dirname(file_path__)):
                os.makedirs(os.path.dirname(file_path__))
            with open(os.path.join(file_path__), "w", encoding="utf-8") as f__:
                f__.write(content__)

        def dump_slot_content(slot_key_, content_type_, content_):
            if isinstance(content_, dict):
                content_ = json.dumps(content_, indent=4)
            if content_type_ == "Scribunto":
                content_type_ = "lua"
            if slot_key_ == "main" and config.skip_slot_suffix_for_main:
                file_name_ = f"{dump_name}.{content_type_}"
            else:
                file_name_ = f"{dump_name}.slot_{slot_key_}.{content_type_}"
            # handle subpages:
            file_path_ = os.path.join(tar_dir, *file_name_.split("/"))
            save_to_file(file_path_, content_)
            if slot_key_ == "main":
                package_page.urlPath = path_prefix + file_name_
            else:
                package_page.slots[slot_key_] = package.PagePackagePageSlot(
                    urlPath=path_prefix + file_name_
                )

        for slot_key in self._slots:
            content = self.get_slot_content(slot_key)
            content_type = self.get_slot_content_model(slot_key)
            dump_slot_content(slot_key, content_type, content)

        # If the slots are empty, we still want files to fill after dumping them
        if config.dump_empty_slots:
            for slot_key in [slot for slot in SLOTS.keys() if slot not in self._slots]:
                content = SLOTS[slot_key]["content_template"]
                content_type = SLOTS[slot_key]["content_model"]
                dump_slot_content(slot_key, content_type, content)

        if self.is_file_page():
            file = self.wtSite._site.images[self.title.split(":")[-1]]
            file_name = f"{page_name}"
            file_path = os.path.join(tar_dir, *file_name.split("/"))  # handle subpages
            with open(file_path, "wb") as fd:
                file.download(fd)
            package_page.fileURLPath = path_prefix + file_name

        return package_page

    @_basemodel_decorator
    class PageUploadConfig(BaseModel):
        pass

    # todo: implement

    def upload(self, config: PageUploadConfig):
        pass



import os
from importlib import reload
from pprint import pprint

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# import src.controller.DeviceType as crtl
# from src.controller.DeviceType import DeviceInstance


# create/update the password file under examples/wiki-admin.pwd
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

do_fetch = True
do_fetch = False

if do_fetch:
    # osw.fetch_schema() #this will load the current entity schema from the OSW instance. You may have to re-run the script to get the updated schema extension. Requires 'pip install datamodel-code-generator'
    osw.fetch_schema(
        OSW.FetchSchemaParam(schema_title="JsonSchema:LIMS/Device/Type", mode="replace")
    )
    osw.fetch_schema(
        OSW.FetchSchemaParam(
            schema_title="JsonSchema:LIMS/Device/Instance", mode="append"
        )
    )
    reload(model)  # only for modules

# check if the schema extension was loaded
try:
    model.Device
except AttributeError:
    print("[Ex] Device not defined")
else:
    print("[Ex] Device defined")

# device = model.Device(manufacturer="Test")
# event = model.Event(start=date.fromisoformat("2022-01-01"))
# entity = model.Entity(label="TestE", extensions=[
#     model.Device(
#        manufacturer = "TestM"
#    ),
#    event
# ])

# pprint(entity)
# osw.store_entity(entity = entity, entity_title="Term:OSW9c64c51bd5fb4162bc1fa9e60468a09d" )

entity2 = osw.load_entity("Term:OSW9c64c51bd5fb4162bc1fa9e60468a09e")
# pprint(entity2)
print(type(entity2))
# ctrl = crtl.DeviceController(dev = entity2)
# ctrl.print()

entity2.print()
entity2.getPlot()
di = model.DeviceInstance(label="test")
di.print()
print(type(di))

# create custom model
# @model._basemodel_decorator
# class MyModel(model.BaseModel, metaclass=OswClassMetaclass, osw_template="MyTemplate", osw_footer_template="MyFooterTemplate"):
#    my_property_2: Optional[str]
#    my_property: Optional[int] =  Field(None, title="My Property")

# print(MyModel.schema_json(indent=4))
# myModel = model.Entity(label="MyModel", extensions=[MyModel(my_property=1)])
# pprint(myModel)
# osw.register_schema(OSW.SchemaRegistration(model_cls=MyModel, schema_name="MyModel"))

# import src.model.LIMS.Device.Type as model2
dt = model.DeviceType(label=[model.Label(label_text="Test")])
# pprint(dt)
# pprint(dt.json())

se = model.DeviceInstance(
    label="PCB",
    statements=[
        model.Statement(predicate="Contains", object="Gold"),
        model.Statement(
            predicate="Contains",
            object="Lead",
            substatements=[model.Statement(predicate="HasConcentration", object="10%")],
        ),
    ],
)
pprint(se.dict())
osw.store_entity(entity=se, entity_title="Term:OSW9c64c51bd5fb4162bc1fa9e60468a09e")



import os
from pprint import pprint

import osw.model.entity as model
from osw.core import OSW
from osw.wtsite import WtSite

# create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
wtsite = WtSite.from_domain("wiki-dev.open-semantic-lab.org", pwd_file_path)
osw = OSW(site=wtsite)

my_entity = model.Item(
    label=[model.Label(text="MyItem")], statements=[model.Statement(predicate="IsA")]
)
pprint(my_entity)

osw.store_entity(my_entity)

my_entity2 = osw.load_entity("Item:" + OSW.get_osw_id(my_entity.uuid))
pprint(my_entity)

osw.delete_entity(my_entity)
